function [H, F, M, report] = fmoRecon_std(G,B,F,M,H,par)
% Solve the deblatting problem
% Find object trajetories, appearance and shape from a single image and
% background
%
% [H, F, M, report] = fmoRecon_std(G,B,F,M,H,par)
%
% input:
% G ... color blurred image MxNxL (L... number of color channels)
% B ... color background image MxNxL (L... number of color channels)
% F ... filters (color foreground objects); cell array { (M1xN1xL) (M2xN2xL) ...}
% M ... masks; cell array in the same format as F; default is { 1 1 ...}
%       cells can contain also 0 or 1. 0... zero mask; 1... get mask from F (F~=0)  
% H ... initial PSFs (trajectories of objects F) MxNxlength(F) (default is
%       empty)
% par . structure containing parameters that override default ones
%
% output:
% H ... estimated PSFs (MxNxlength(F))
% F ... estimated foreground objects (object appearance)
% M ... estimated masks (object shape)
% report ... structure; 
%            field E: energy in every iteration; 
%            field PAR: parameters of the algorithm
%

%% DEFAULT PARAMETERS
% Number of iterations in each step
maxiter_main = 100; % max number of iterations in the main loop
maxiter_fm = 0; % number of initial minF and minM iterations
maxiter_H = 1; % max number of iterations in the minHstep
maxiter_F = 1; % max number of iterations in the minFMstep

VERBOSE = 1;
%% MAIN PARAMETERS
%(alpha, beta, xi, maxbeta, tau, gamma)
% 1e1 1e1 1e1 1e3 1.1 1e0
% 1e1 1e2 1e2 1e4 1.1 1e1
Lp = 1; % type of sparse norm l_p
%alpha = 1e1; %1; % sparse weight
beta = 1e3; %1e2; %1e3; %1e1; % beta || a - v ||^2
alphaF = 1e-3; % 1e-2; % 1e-4;
betaF = 1e-1; %1e-1; %1e0; %1e-2; 
alphaM = 0.5*betaF;
betaM = betaF;
betaFM = 1e-1; %1e4; %1e2;
gamma = 1e0; %1e0 for pinkpong; %1e-4 was for image intensities 0..255 % data term weight

%alphaF = 0; % 1e-2; % 1e-4;
%betaF = alphaF*100; %1e-1; %1e0; %1e-2; 
%alphaM = 1e-2; %1e-2;
%betaM = alphaM*100;
%betaFM = 0; %1e4; %1e2;
%gamma = 1; %

reltol = 1e-4; %1e-4; % CG reltol

%% DEFAULT PARAMETERS END HERE

if exist('par','var') && ischar(par)
    [pathstr,name] = fileparts(par);
    curf = pwd;
    if ~isempty(pathstr)
        cd(pathstr);
    end
    eval(name);
    cd(curf);
end

if nargin < 4 || isempty(M)
    flag = num2cell(ones(1,length(F)));
else
    flag = M;
end

gsize = [size(G,1),size(G,2),size(G,3)];
L = gsize(3); % number of color channels

if VERBOSE>1
    FIG_HANDLE_F = figure(1);
    FIG_HANDLE_H = figure(2);
    %FIG_HANDLE_M = figure(3);
end


% variables used in min_H
Hsize = [gsize(1),gsize(2),length(F)];
hsize = [gsize(1)*gsize(2),length(F)];
Tsize = [Hsize, L];
tsize = [hsize, L];
Fsize = [size(F{1},1), size(F{1},2), length(F), L];
Msize = Fsize(1:3);
%HTH = 0;
% shift H as we use fft2 instead of conv to perform convolution
H = circshift(H,-floor(Msize(1:2)/2));
if nargin < 5 || isempty(H)
    H = zeros(Hsize);
    FH = zeros(hsize);
else
    FH = reshape(fft2(H),hsize);
end
V = zeros(Hsize); % Auxiliar variable
A = zeros(Hsize); % Lagrange multiplier
%Prh = asetupLnormPrior(Lp,alpha,beta); % Lp norm
%h_projection = @(x) (Prh.fh(x,abs(x)));
h_projection = @SimplexProj; % projection to a simplex
V = h_projection(H-A);


% variables used in min_F
Dx = [1 -1; 0 0];
Dy = [1 0; -1 0];
%DTD = kron(eye(Fsize(3)*Fsize(4)), ...
%    fftconvcirc2matrix(padarray(Dx,Fsize(1:2),'post'),Fsize(1:2)) + ...
%    fftconvcirc2matrix(padarray(Dy,Fsize(1:2),'post'),Fsize(1:2)));
DTD2 = conv2(Dx,rot90(Dx,2),'full') + conv2(Dy,rot90(Dy,2),'full');
W = zeros(Fsize); % Auxiliar variable
C = zeros(Fsize); % Lagrange multiplier
Dsize = Fsize; Dsize(1:2) = Dsize(1:2)+1;
Wx = zeros(Dsize);
Cx = zeros(Dsize);
Wy = zeros(Dsize);
Cy = zeros(Dsize);

D = zeros(Msize); % Lagrange multiplier
Z = zeros(Msize); % Auxiliar variable

Zx = zeros(Dsize(1:3));
DDx = zeros(Dsize(1:3));
Zy = zeros(Dsize(1:3));
DDy = zeros(Dsize(1:3));

for c = 1:hsize(2)
    %FF(:,c,:) = reshape(fft2(double(F{c}),gsize(1),gsize(2)),[hsize(1), 1, L]);
    if numel(flag{c}) ==  1      
        if flag{c} == 1 %initialize mask with ones or as delta
            M{c} = double(any(F{c}~=0,3));
        else
            M{c} = ones(size(F{c},1),size(F{c},2));
        end
    else %initialize mask with the input argument
        M{c} = flag{c}; 
    end
    %FM(:,c,:) = reshape(fft2(double(M{c}),gsize(1),gsize(2)),[hsize(1), 1, L]);
end
M = reshape(cell2mat(M),Msize); % replicate mask for every color channel
F = reshape(cell2mat(F),Fsize);

% set F to delta
%F = zeros(Fsize);
%F(floor(Fsize(1)/2)+1, floor(Fsize(2)/2)+1,:, :) = 1;

% set F to ones
%F = zeros(Fsize); F(:,:,:,3) = 1;

%PrF = asetupLnormPrior(Lp,alphaF,betaF);
%W = PrF.fh(F-C,abs(F-C));
W = F; W(W<0) = 0;
Z = M; Z(Z<0) = 0; Z(Z>1) = 1;
FF = reshape(fft2(F,gsize(1),gsize(2)),tsize);
FM = reshape(fft2(M,gsize(1),gsize(2)),hsize);

% common input variables
FGB = reshape(fft2(G-B),[hsize(1),1,L]);
%rB = B(Fsize(1):end,Fsize(2):end,:);
%rGB = G(Fsize(1):end,Fsize(2):end,:)-rB;


report.E = [];

fh_projection = @dyksproj2Set;
% initial estimation of F
for j = 1:maxiter_fm
    minFMstep;
end

report.PAR = struct('Lp', Lp, 'alphaF', alphaF, 'alphaM', alphaM, ...
    'beta', beta, 'betaF', betaF, 'betaM', betaM, 'betaFM', betaFM, 'gamma', gamma, 'reltol', reltol);
report.E = [report.E; calcE];

for i_main = 1:maxiter_main
    if VERBOSE
        disp([num2str(i_main)]);
    end
    minFMstep;
    
    minHstep;
    
    report.E = [report.E; calcE];
end

%shift H back for use with conv instead of fft2
H = circshift(H,floor(Msize(1:2)/2));



%% END

%% update H (object trajectory)
function minHstep

Fb1 = sum(conj(FF).*FGB,3) - ...
        sum(conj(FM).*reshape(fft2(B.*(G-B)),[hsize(1),1,L]),3);
b1 = real(ifft2(reshape(Fb1,Hsize)));
% additive
%Pr = asetupLnormPrior(Lp,alpha,beta); 

for i=1:maxiter_H
    % size hsize 
    b = b1 + beta/gamma*(V+A);
    %b = Fb1 + beta/gamma*reshape(fft2(V+A),hsize);
    
    [xmin,fl,relres,iter,resvec] = mycg(@gradH,vec(b),reltol,1000,[],vec(H));
    if VERBOSE
        disp(['minH ',num2str(i),' - flag, iter:',num2str([fl iter])]);
    end
    H = reshape(xmin,Hsize);
    %FH = reshape(xmin,hsize);
    %H = real(ifft2(reshape(FH,Hsize)));
    
    %% additive
    %V = Pr.fh(H-A,abs(H-A)); % Lp norm
    V = h_projection(H-A);
    V(V<0) = 0;
    % update Bregman variable (Lagrange variable)
    A = A + V - H;
    if VERBOSE>1
        update4Fig(FIG_HANDLE_H,{'V', reshape(V,Hsize(1),[])},{['H, sum=',num2str(sum(H(:)),2)], reshape(H,Hsize(1),[])},{'A', reshape(A,Hsize(1),[])},...
            {'G-B',linscale(G-B)} );   
    end
end
FH = reshape(fft2(H),hsize);
%report.E = [report.E, calcE];
end

function r = gradH(x)
    %X = reshape(x,hsize);
    X = reshape(fft2(reshape(x,Hsize)),hsize);
    % (F - diag(B) M)
    FT = reshape(sum(FF.*X,2),[gsize(1),gsize(2),L]) - ...
        fft2(B.*real(ifft2(reshape(sum(FM.*X,2),[gsize(1),gsize(2)]))));
    T = real(ifft2(FT));
    % (F^T - M^T diag(B))
    FT = reshape(FT,[hsize(1) 1 L]);
    T = sum(conj(FF).*FT,3) - ...
        sum(conj(FM).*reshape(fft2(B.*T),[hsize(1),1,L]),3);
    % beta/gamma I
    %r = vec(T) + beta/gamma*x;
    r = vec(real(ifft2(reshape(T,Hsize)))) + beta/gamma*x;
end

%% update F and M simultenously
function minFMstep
b1 = conj(FH).*FGB;
b1 = real(ifft2(reshape(b1,[gsize(1), gsize(2), hsize(2), L])));
b1 = b1(1:Fsize(1),1:Fsize(2),:,:);
b2 = conj(FH).*sum(reshape(fft2(B.*(G-B)),size(FGB)),3);
b2 = real(ifft2(reshape(b2,[gsize(1), gsize(2), hsize(2)])));
b2 = -b2(1:Fsize(1),1:Fsize(2),:);
Pr = asetupLnormPrior(Lp,alphaF,betaF); 
PrM = asetupLnormPrior(Lp,alphaM,betaM);
for i=1:maxiter_F
    ba = b1 + betaFM/gamma*(W+C) + ...
        betaF/gamma*(convn(Wx+Cx, rot90(Dx,2),'valid') + convn(Wy+Cy, rot90(Dy,2),'valid'));
    bb = b2 + betaFM/gamma*(Z+D) + ...
        betaM/gamma*(convn(Zx+DDx, rot90(Dx,2),'valid') + convn(Zy+DDy, rot90(Dy,2),'valid'));
    [xmin,fl,relres,iter,resvec] = mycg(@gradFM,[ba(:);bb(:)],reltol,100,[],[F(:);M(:)]);
    
    F = reshape(xmin(1:prod(Fsize)),Fsize);
    M = reshape(xmin(prod(Fsize)+1:end),Msize);
    
    Z = M-D;
    W = F-C;
    [T, projiter] = fh_projection([Z(:), reshape(W,[],L)].');
    
    Z(:) = T(1,:);
    W(:) = T(2:L+1,:).';
       
    xD = convn(F,Dx,'full');
    yD = convn(F,Dy,'full');
    xDm = xD - Cx;
    yDm = yD - Cy;
    nDm = repmat(sqrt(sum(xDm.^2,4) + sum(yDm.^2,4)),[1 1 1 Fsize(4)]);
    %nDm = sqrt(xDm.^2 + yDm.^2);
    Wy = Pr.fh(yDm,nDm);
    Wx = Pr.fh(xDm,nDm);
 
    xM = convn(M,Dx,'full');
    yM = convn(M,Dy,'full');
    xDm = xM - DDx;
    yDm = yM - DDy;
    nDm = sqrt(sum(xDm.^2,4) + sum(yDm.^2,4));
    Zy = PrM.fh(yDm,nDm);
    Zx = PrM.fh(xDm,nDm);
    
    % update Bregman variables (Lagrange variables)
    D = D - M + Z;
    C = C + W - F;
    Cx = Cx + Wx - xD;
    Cy = Cy + Wy - yD;
    
    DDx = DDx + Zx - xM;
    DDy = DDy + Zy - yM;
    
    if VERBOSE
        disp(['minFM ',num2str(i),' - flag, iter, proj_iter: ',num2str([fl iter projiter])]);
    end
    if VERBOSE>1
        update4Fig(FIG_HANDLE_F,{'Z', linscale(reshape(Z,Fsize(1),[]))},...
            {'F', linscale(reshape(F,Fsize(1),[],L))},...
            {'M', linscale(reshape(M,Fsize(1),[]))},...
            {'W', linscale(reshape(W,size(W,1),[],L))});     
    end
end
FF = reshape(fft2(F,gsize(1),gsize(2)),tsize);
FM = reshape(fft2(M,gsize(1),gsize(2)),hsize);
end
function r = gradFM(x)
    Xf = reshape(x(1:prod(Fsize)),Fsize);
    Xm = reshape(x(prod(Fsize)+1:end),Msize);
    FXf = reshape(fft2(Xf,Hsize(1),Hsize(2)),tsize);
    FXm = reshape(fft2(Xm,Hsize(1),Hsize(2)),tsize(1:2));
    T = sum(FH.*FXf,2) - reshape(fft2(B.*real(ifft2(reshape(sum(FH.*FXm,2),Hsize(1:2))))),size(FGB));
    Tf = real(ifft2(reshape(conj(FH).*T,Tsize)));
    Tf = Tf(1:Fsize(1),1:Fsize(2),:,:);
    Tm = conj(FH).*reshape(fft2(sum(B.*real(ifft2(reshape(T,gsize))),3)),[hsize(1),1]);
    Tm = real(ifft2(reshape(Tm,[gsize(1), gsize(2), hsize(2)])));
    Tm = -Tm(1:Fsize(1),1:Fsize(2),:);
    r = [vec(Tf + betaFM/gamma*Xf + betaF/gamma*convn(Xf,DTD2,'same'));...
        vec(Tm + betaFM/gamma*Xm + betaM/gamma*convn(Xm,DTD2,'same'))];
end

function E = calcE
    FE = reshape(sum(FF.*repmat(FH,1,1,L),2),[gsize(1),gsize(2),L]) - ...
           fft2(B.*real(ifft2(reshape(sum(FM.*repmat(FH,1,1,L),2),[gsize(1),gsize(2),L])))) - ...
           reshape(FGB,[gsize(1),gsize(2),L]);
    E = sum(abs(FE(:)).^2)/numel(FE);
end

end

