function [ncorr, h_shifted, shift] = psfncorr(h, gt)
% PSFNCORR returns normalized correlation between ground-truth and estimated PSFs, 
% up to arbitrary (as in 'full' in conv2) whole pixel translation. 
% (ie - the best possible match, translation-wise)
% Both psfs must be prenormalized to sum to 1, otherwise the result may not make sense (be even negative)
%
% h_shifted - correctly shifted (and cropped) h to best match gt
% shift - respective shift of 'h' to 'gt'. Positive values: 'h' shifted "to the right" of 'gt'.
%
% last changed: 01/2020

% pad 'h' to accomodate all transaltions of 'gt' so that 'valid' convolution in the padded 'h' corresponds to 'full' convolution in the unpadded. (this is so that the CC is calculated in the whole area of 'gt' and 'h', not only in the interecting part in case of full conv)
sz_gt = size(gt);
sz_h = size(h);
gt_padded = padarray(gt, sz_h - 1, 0, 'both'); % 'sliding' window
h_padded = padarray(h, sz_gt + sz_h - 2, 0, 'both'); % 'fixed' window

% calculate correlation
t1 = filter2(ones(size(gt_padded)), h_padded.^2, 'valid'); % corresponds to 'full' conv of unpadded h/gt but with all pixels 'included'
t2 = filter2(gt_padded, h_padded, 'valid');
t3 = sum(gt(:).^2);
ssd = t2./(sqrt(t1).*sqrt(t3));
%ssd = t1 - 2*t2 + t3; % ssd value in all shift locations

% return max correlation
ncorr = max(ssd(:));
%min_ssd = min(ssd(:));
%psnr = 10*log10(1/min_ssd); % max possible error is 1 (delta against infinitely spread-out PSF)

if(nargout >= 2)
	[r, c] = find(ssd == ncorr); % row/col of min locations
	zero = (sz_h + sz_gt)/2; % row/col corresponding to zero shift (because of all the padding, valid convolution behaves like 'full')

	% in case of several minima, take the one with smallest shift
	[~, idx] = min((r-zero(1)).^2+(c-zero(2)).^2);
	shift = [r(idx) c(idx)];

	% cut the corresponding region from h
	pos = shift + sz_h - 1; % topleft corner in h_padded
	h_shifted = h_padded(pos(1):pos(1)+sz_gt(1)-1, pos(2):pos(2)+sz_gt(2)-1);
	h_shifted = h_shifted/sum(h_shifted(:)); % sum to 1

	% recalculate to relative shift
	shift = shift - sz_gt;
end
end
